@@ -20,20 +20,41 @@ SEXP fns_validate_unique_names = NULL;
 SEXP vctrs_as_minimal_names(SEXP names);
 SEXP vec_as_universal_names(SEXP names, bool quiet);
 SEXP vec_validate_unique_names(SEXP names);
+SEXP vec_as_custom_names(SEXP names, const struct name_repair_opts* opts);
+static void vec_validate_minimal_names(SEXP names, R_len_t n);
 
 
-// [[ include("vctrs.h") ]]
-SEXP vec_as_names(SEXP names, enum name_repair_arg type, bool quiet) {
-  switch (type) {
+// [[ include("names.h") ]]
+SEXP vec_as_names(SEXP names, const struct name_repair_opts* opts) {
+  if (!opts) {
+    return names;
+  }
+  switch (opts->type) {
   case name_repair_none: return names;
   case name_repair_minimal: return vctrs_as_minimal_names(names);
-  case name_repair_unique: return vec_as_unique_names(names, quiet);
-  case name_repair_universal: return vec_as_universal_names(names, quiet);
+  case name_repair_unique: return vec_as_unique_names(names, opts->quiet);
+  case name_repair_universal: return vec_as_universal_names(names, opts->quiet);
   case name_repair_check_unique: return vec_validate_unique_names(names);
+  case name_repair_custom: return vec_as_custom_names(names, opts);
   }
   never_reached("vec_as_names");
 }
 
+// [[ register() ]]
+SEXP vctrs_as_names(SEXP names, SEXP repair, SEXP quiet) {
+  if (!r_is_bool(quiet)) {
+    Rf_errorcall(R_NilValue, "`quiet` must a boolean value.");
+  }
+
+  struct name_repair_opts repair_opts = new_name_repair_opts(repair, quiet);
+  PROTECT_NAME_REPAIR_OPTS(&repair_opts);
+
+  SEXP out = vec_as_names(names, &repair_opts);
+
+  UNPROTECT(1);
+  return out;
+}
+
 SEXP vec_as_universal_names(SEXP names, bool quiet) {
   SEXP quiet_obj = PROTECT(r_lgl(quiet));
   SEXP out = vctrs_dispatch2(syms_as_universal_names, fns_as_universal_names,
@@ -53,6 +74,26 @@ SEXP vec_validate_unique_names(SEXP names) {
   return out;
 }
 
+SEXP vec_as_custom_names(SEXP names, const struct name_repair_opts* opts) {
+  names = PROTECT(vctrs_as_minimal_names(names));
+
+  // Don't use vctrs dispatch utils because we match argument positionally
+  SEXP call = PROTECT(Rf_lang2(syms_repair, syms_names));
+  SEXP mask = PROTECT(r_new_environment(R_GlobalEnv, 2));
+  Rf_defineVar(syms_repair, opts->fn, mask);
+  Rf_defineVar(syms_names, names, mask);
+  SEXP out = PROTECT(Rf_eval(call, mask));
+
+  vec_validate_minimal_names(out, Rf_length(names));
+
+  if (!opts->quiet) {
+    describe_repair(names, out);
+  }
+
+  UNPROTECT(4);
+  return out;
+}
+
 
 // [[ register(); include("vctrs.h") ]]
 SEXP vec_names(SEXP x) {
@@ -364,7 +405,7 @@ SEXP vctrs_unique_names(SEXP x, SEXP quiet) {
 
 // [[ include("utils.h") ]]
 SEXP vec_unique_names(SEXP x, bool quiet) {
-  SEXP names = PROTECT(Rf_getAttrib(x, R_NamesSymbol));
+  SEXP names = PROTECT(vec_names(x));
   SEXP out = vec_unique_names_impl(names, vec_size(x), quiet);
   UNPROTECT(1);
   return out;
@@ -417,9 +458,6 @@ static void describe_repair(SEXP old_names, SEXP new_names) {
 }
 
 
-static SEXP outer_names_cat(const char* outer, SEXP names);
-static SEXP outer_names_seq(const char* outer, R_len_t n);
-
 // [[ register() ]]
 SEXP vctrs_outer_names(SEXP names, SEXP outer, SEXP n) {
   if (names != R_NilValue && TYPEOF(names) != STRSXP) {
@@ -453,10 +491,10 @@ SEXP outer_names(SEXP names, SEXP outer, R_len_t n) {
     if (n == 1) {
       return r_str_as_character(outer);
     } else {
-      return outer_names_seq(CHAR(outer), n);
+      return r_seq_chr(CHAR(outer), n);
     }
   } else {
-    return outer_names_cat(CHAR(outer), names);
+    return r_chr_paste_prefix(names, CHAR(outer), "..");
   }
 }
 
@@ -539,23 +577,27 @@ static SEXP glue_as_name_spec(SEXP spec) {
 }
 
 
-static SEXP outer_names_cat(const char* outer, SEXP names) {
+// [[ include("names.h") ]]
+SEXP r_chr_paste_prefix(SEXP names, const char* prefix, const char* sep) {
   names = PROTECT(Rf_shallow_duplicate(names));
   R_len_t n = Rf_length(names);
 
-  int outer_len = strlen(outer);
+  int outer_len = strlen(prefix);
   int names_len = r_chr_max_len(names);
 
-  int total_len = outer_len + names_len + strlen("..") + 1;
+  int sep_len = strlen(sep);
+  int total_len = outer_len + names_len + sep_len + 1;
 
   R_CheckStack2(total_len);
   char buf[total_len];
   buf[total_len - 1] = '\0';
   char* bufp = buf;
 
-  memcpy(bufp, outer, outer_len); bufp += outer_len;
-  *bufp = '.'; bufp += 1;
-  *bufp = '.'; bufp += 1;
+  memcpy(bufp, prefix, outer_len); bufp += outer_len;
+
+  for (int i = 0; i < sep_len; ++i) {
+    *bufp++ = sep[i];
+  }
 
   SEXP* p = STRING_PTR(names);
 
@@ -573,77 +615,255 @@ static SEXP outer_names_cat(const char* outer, SEXP names) {
   return names;
 }
 
-static SEXP outer_names_seq(const char* outer, R_len_t n) {
-  int total_len = 24 + strlen(outer) + 1;
+// [[ include("names.h") ]]
+SEXP r_seq_chr(const char* prefix, R_len_t n) {
+  int total_len = 24 + strlen(prefix) + 1;
 
   R_CheckStack2(total_len);
   char buf[total_len];
 
-  return r_chr_iota(n, buf, total_len, outer);
+  return r_chr_iota(n, buf, total_len, prefix);
 }
 
 
 // Initialised at load time
-SEXP syms_set_rownames = NULL;
-SEXP fns_set_rownames = NULL;
+SEXP syms_set_rownames_fallback = NULL;
+SEXP fns_set_rownames_fallback = NULL;
 
-// [[ include("utils.h") ]]
-SEXP set_rownames(SEXP x, SEXP names) {
-  return vctrs_dispatch2(syms_set_rownames, fns_set_rownames,
+static SEXP set_rownames_fallback(SEXP x, SEXP names) {
+  return vctrs_dispatch2(syms_set_rownames_fallback, fns_set_rownames_fallback,
                          syms_x, x,
                          syms_names, names);
 }
 
+// Initialised at load time
+SEXP syms_set_names_fallback = NULL;
+SEXP fns_set_names_fallback = NULL;
 
-enum name_repair_arg validate_name_repair(SEXP arg) {
-  if (!Rf_length(arg)) {
-    Rf_errorcall(R_NilValue, "`.name_repair` must be a string. See `?vctrs::vec_as_names`.");
+static SEXP set_names_fallback(SEXP x, SEXP names) {
+  return vctrs_dispatch2(syms_set_names_fallback, fns_set_names_fallback,
+                         syms_x, x,
+                         syms_names, names);
+}
+
+static void check_names(SEXP x, SEXP names) {
+  if (names == R_NilValue) {
+    return;
   }
 
-  arg = r_chr_get(arg, 0);
-
-  if (arg == strings_none) {
-    return name_repair_none;
-  }
-  if (arg == strings_minimal) {
-    return name_repair_minimal;
-  }
-  if (arg == strings_unique) {
-    return name_repair_unique;
-  }
-  if (arg == strings_universal) {
-    return name_repair_universal;
-  }
-  if (arg == strings_check_unique) {
-    return name_repair_check_unique;
+  if (TYPEOF(names) != STRSXP) {
+    Rf_errorcall(
+      R_NilValue,
+      "`names` must be a character vector, not a %s.",
+      Rf_type2char(TYPEOF(names))
+    );
   }
 
-  Rf_errorcall(R_NilValue, "`.name_repair` can't be \"%s\". See `?vctrs::vec_as_names`.", CHAR(arg));
+  R_len_t x_size = vec_size(x);
+  R_len_t names_size = vec_size(names);
+
+  if (x_size != names_size) {
+    Rf_errorcall(
+      R_NilValue,
+      "The size of `names`, %i, must be the same as the size of `x`, %i.",
+      names_size,
+      x_size
+    );
+  }
+}
+
+SEXP vec_set_rownames(SEXP x, SEXP names) {
+  if (OBJECT(x)) {
+    return set_rownames_fallback(x, names);
+  }
+
+  int nprot = 0;
+
+  SEXP dim_names = Rf_getAttrib(x, R_DimNamesSymbol);
+
+  // Early exit when no new row names and no existing row names
+  if (names == R_NilValue) {
+    if (dim_names == R_NilValue || VECTOR_ELT(dim_names, 0) == R_NilValue) {
+      return x;
+    }
+  }
+
+  x = PROTECT_N(r_maybe_duplicate(x), &nprot);
+
+  if (dim_names == R_NilValue) {
+    dim_names = PROTECT_N(Rf_allocVector(VECSXP, vec_dim_n(x)), &nprot);
+  }
+
+  SET_VECTOR_ELT(dim_names, 0, names);
+
+  Rf_setAttrib(x, R_DimNamesSymbol, dim_names);
+
+  UNPROTECT(nprot);
+  return x;
+}
+
+// FIXME: Do we need to get the vec_proxy() and only fall back if it doesn't
+// exist? See #526 and #531 for discussion and the related issue.
+// [[ include("utils.h"); register() ]]
+SEXP vec_set_names(SEXP x, SEXP names) {
+  // Never on a data frame
+  if (is_data_frame(x)) {
+    return x;
+  }
+
+  check_names(x, names);
+
+  if (has_dim(x)) {
+    return vec_set_rownames(x, names);
+  }
+
+  if (OBJECT(x)) {
+    return set_names_fallback(x, names);
+  }
+
+  // Early exit if no new names and no existing names
+  if (names == R_NilValue && Rf_getAttrib(x, R_NamesSymbol) == R_NilValue) {
+    return x;
+  }
+
+  x = PROTECT(r_maybe_duplicate(x));
+
+  Rf_setAttrib(x, R_NamesSymbol, names);
+
+  UNPROTECT(1);
+  return x;
+}
+
+SEXP vctrs_validate_name_repair_arg(SEXP arg) {
+  struct name_repair_opts opts = new_name_repair_opts(arg, true);
+  if (opts.type == name_repair_custom) {
+    return opts.fn;
+  } else if (Rf_length(arg) != 1) {
+    return r_str_as_character(r_str(name_repair_arg_as_c_string(opts.type)));
+  } else {
+    return arg;
+  }
+}
+
+void stop_name_repair() {
+  Rf_errorcall(R_NilValue, "`.name_repair` must be a string or a function. See `?vctrs::vec_as_names`.");
+}
+
+struct name_repair_opts new_name_repair_opts(SEXP name_repair, bool quiet) {
+  struct name_repair_opts opts = {
+    .type = 0,
+    .fn = R_NilValue,
+    .quiet = quiet
+  };
+
+  switch (TYPEOF(name_repair)) {
+  case STRSXP: {
+    if (!Rf_length(name_repair)) {
+      stop_name_repair();
+    }
+
+    SEXP c = r_chr_get(name_repair, 0);
+
+    if (c == strings_none) {
+      opts.type = name_repair_none;
+    } else if (c == strings_minimal) {
+      opts.type = name_repair_minimal;
+    } else if (c == strings_unique) {
+      opts.type = name_repair_unique;
+    } else if (c == strings_universal) {
+      opts.type = name_repair_universal;
+    } else if (c == strings_check_unique) {
+      opts.type = name_repair_check_unique;
+    } else {
+      Rf_errorcall(R_NilValue, "`.name_repair` can't be \"%s\". See `?vctrs::vec_as_names`.", CHAR(name_repair));
+    }
+
+    return opts;
+  }
+
+  case LANGSXP:
+    opts.fn = r_as_function(name_repair, ".name_repair");
+    opts.type = name_repair_custom;
+    return opts;
+
+  case CLOSXP:
+    opts.fn = name_repair;
+    opts.type = name_repair_custom;
+    return opts;
+
+  default:
+    stop_name_repair();
+  }
+
+  never_reached("new_name_repair_opts");
 }
 
 // [[ include("vctrs.h") ]]
-const char* name_repair_arg_as_c_string(enum name_repair_arg arg) {
-  switch (arg) {
+const char* name_repair_arg_as_c_string(enum name_repair_type type) {
+  switch (type) {
   case name_repair_none: return "none";
   case name_repair_minimal: return "minimal";
   case name_repair_unique: return "unique";
   case name_repair_universal: return "universal";
   case name_repair_check_unique: return "check_unique";
+  case name_repair_custom: return "custom";
   }
   never_reached("name_repair_arg_as_c_string");
 }
 
+static void vec_validate_minimal_names(SEXP names, R_len_t n) {
+  if (names == R_NilValue) {
+    Rf_errorcall(R_NilValue, "Names repair functions can't return `NULL`.");
+  }
+
+  if (TYPEOF(names) != STRSXP) {
+    Rf_errorcall(R_NilValue, "Names repair functions must return a character vector.");
+  }
+
+  if (n >= 0 && Rf_length(names) != n) {
+    Rf_errorcall(R_NilValue,
+                 "Repaired names have length %d instead of length %d.",
+                 Rf_length(names),
+                 n);
+  }
+
+  if (r_chr_has_string(names, NA_STRING)) {
+    Rf_errorcall(R_NilValue, "Names repair functions can't return `NA` values.");
+  }
+}
+SEXP vctrs_validate_minimal_names(SEXP names, SEXP n_) {
+  R_len_t n = -1;
+
+  if (TYPEOF(n_) == INTSXP) {
+    if (Rf_length(n_) != 1) {
+      Rf_error("Internal error (minimal names validation): `n` must be a single number.");
+    }
+    n = INTEGER(n_)[0];
+  }
+
+  vec_validate_minimal_names(names, n);
+  return names;
+}
+
+
+struct name_repair_opts default_unique_repair_opts_obj;
 
 void vctrs_init_names(SEXP ns) {
-  syms_set_rownames = Rf_install("set_rownames");
+  syms_set_rownames_fallback = Rf_install("set_rownames_fallback");
+  syms_set_names_fallback = Rf_install("set_names_fallback");
   syms_as_universal_names = Rf_install("as_universal_names");
   syms_validate_unique_names = Rf_install("validate_unique");
 
-  fns_set_rownames = r_env_get(ns, syms_set_rownames);
+  fns_set_rownames_fallback = r_env_get(ns, syms_set_rownames_fallback);
+  fns_set_names_fallback = r_env_get(ns, syms_set_names_fallback);
   fns_as_universal_names = r_env_get(ns, syms_as_universal_names);
   fns_validate_unique_names = r_env_get(ns, syms_validate_unique_names);
 
   syms_glue_as_name_spec = Rf_install("glue_as_name_spec");
   fns_glue_as_name_spec = r_env_get(ns, syms_glue_as_name_spec);
   syms_internal_spec = Rf_install("_spec");
+
+  default_unique_repair_opts_obj.type = name_repair_unique;
+  default_unique_repair_opts_obj.fn = R_NilValue;
+  default_unique_repair_opts_obj.quiet = false;
 }
