@@ -1,13 +1,5 @@
 context("test-names")
 
-
-# API -----------------------------------------------------------------
-
-test_that("Consistent values for repair argument", {
-  expect_identical(formals(validate_repair)$repair, formals(vec_as_names)$repair)
-  expect_identical(formals(validate_repair)$repair, formals(vec_names2)$repair)
-})
-
 # vec_names() ---------------------------------------------------------
 
 test_that("vec_names() retrieves names", {
@@ -20,7 +12,7 @@ test_that("vec_names() retrieves names", {
 })
 
 test_that("vec_names() dispatches", {
-  scoped_global_bindings(
+  local_methods(
     names.vctrs_foobar = function(x) "dispatched!"
   )
   expect_identical(vec_names(foobar()), "dispatched!")
@@ -109,7 +101,11 @@ test_that("vec_as_names() keeps the names of a named vector", {
 })
 
 test_that("vec_as_names() accepts and checks repair function", {
-  expect_identical(vec_as_names(c("", ""), repair = ~ rep_along(.x, "foo")), c("foo", "foo"))
+  f <- local({
+    local_obj <- "foo"
+    ~ rep_along(.x, local_obj)
+  })
+  expect_identical(vec_as_names(c("", ""), repair = f), c("foo", "foo"))
   expect_error(vec_as_names(c("", ""), repair = function(nms) "foo"), "length 1 instead of length 2")
 })
 
@@ -117,10 +113,10 @@ test_that("vec_as_names() repairs names before invoking repair function", {
   expect_identical(vec_as_names(chr(NA, NA), repair = identity), c("", ""))
 })
 
-test_that("validate_minimal() checks names", {
-  expect_error(validate_minimal(1), "must return a character vector")
-  expect_error(validate_minimal(NULL), "can't return `NULL`")
-  expect_error(validate_minimal(chr(NA)), "can't return `NA` values")
+test_that("validate_minimal_names() checks names", {
+  expect_error(validate_minimal_names(1), "must return a character vector")
+  expect_error(validate_minimal_names(NULL), "can't return `NULL`")
+  expect_error(validate_minimal_names(chr(NA)), "can't return `NA` values")
 })
 
 test_that("validate_unique() checks unique names", {
@@ -131,6 +127,17 @@ test_that("validate_unique() checks unique names", {
   expect_error(validate_unique(chr("...")), class = "vctrs_error_names_cannot_be_dot_dot")
 })
 
+test_that("vec_as_names_validate() validates repair arguments", {
+  expect_identical(
+    validate_name_repair_arg(c("unique", "check_unique")),
+    "unique"
+  )
+  expect_identical(
+    validate_name_repair_arg(~ toupper(.))(letters),
+    LETTERS
+  )
+})
+
 
 # vec_repair_names() -------------------------------------------------------
 
@@ -147,6 +154,87 @@ test_that("vec_repair_names() handles data frames and arrays", {
   expect_identical(row.names(vec_repair_names(as.matrix(df), "unique")), c("...1", "...2"))
 })
 
+# vec_set_names() -----------------------------------------------------------
+
+test_that("vec_set_names() sets atomic names", {
+  x <- 1:2
+  names <- c("x1", "x2")
+  exp <- set_names(x, names)
+  expect_equal(vec_set_names(x, names), exp)
+})
+
+test_that("vec_set_names() sets matrix/array names", {
+  x <- matrix(1:2)
+  names <- c("x1", "x2")
+  exp <- x
+  rownames(exp) <- names
+  expect_equal(vec_set_names(x, names), exp)
+
+  y <- array(1:4, dim = c(2, 1, 2))
+  exp <- y
+  rownames(exp) <- names
+  expect_equal(vec_set_names(y, names), exp)
+})
+
+test_that("vec_set_names() does not set row names on data frames", {
+  x <- data.frame(a = 1, b = 2)
+  expect_equal(vec_set_names(x, "r1"), x)
+})
+
+test_that("vec_set_names() leaves existing data frame row names intact", {
+  x <- data.frame(a = 1, b = 2, row.names = "original")
+  expect_equal(rownames(vec_set_names(x, "new")), "original")
+})
+
+test_that("vec_set_names() correctly sets names on POSIXlt objects", {
+  x <- as.POSIXlt(new_datetime(0))
+  exp <- set_names(x, "a")
+  expect_equal(vec_set_names(x, "a"), exp)
+})
+
+test_that("vec_set_names() falls back to `names<-` with proxied objects", {
+  x <- structure(1, class = "foobar")
+  exp <- set_names(x, "a")
+
+  expect_equal(vec_set_names(x, "a"), exp)
+
+  local_methods(`names<-.foobar` = function(x, value) "fallback!")
+  expect_equal(vec_set_names(x, "a"), "fallback!")
+})
+
+test_that("vec_set_names() falls back to `rownames<-` with shaped proxied objects", {
+  x <- structure(1:2, dim = c(2L, 1L), class = "foobar")
+  names <- c("r1", "r2")
+  exp <- x
+  rownames(exp) <- names
+
+  expect_equal(vec_set_names(x, names), exp)
+
+  # `rownames<-` is not generic, but eventually calls `dimnames<-` which is
+  local_methods(`dimnames<-.foobar` = function(x, value) "fallback!")
+  expect_equal(vec_set_names(x, names), "fallback!")
+})
+
+test_that("vec_set_names() can set NULL names", {
+  x <- 1:2
+  expect_equal(vec_set_names(x, NULL), x)
+
+  x_named <- set_names(x)
+  expect_equal(vec_set_names(x_named, NULL), x)
+
+  x_mat <- as.matrix(x)
+  expect_equal(vec_set_names(x_mat, NULL), x_mat)
+
+  x_mat_named <- x_mat
+  rownames(x_mat_named) <- c("1", "2")
+  exp <- matrix(x_mat, dimnames = list(NULL, NULL))
+  expect_equal(vec_set_names(x_mat_named, NULL), exp)
+})
+
+test_that("vec_set_names() errors with bad `names`", {
+  expect_error(vec_set_names(1, 1), "character vector, not a double")
+  expect_error(vec_set_names(1, c("x", "y")), "The size of `names`, 2")
+})
 
 # minimal names -------------------------------------------------------------
 
